from collections import defaultdict
from lib2to3.fixer_base import BaseFix as lib2to3_BaseFix
import lib2to3.fixer_util

from my2to3.trace import create_table, get_data, tracing_functions


class BaseFix(lib2to3_BaseFix):
  pass


class BaseStaticTraceFix(BaseFix):
  def __init__(self, *args, **kwargs):
    super(BaseStaticTraceFix, self).__init__(*args, **kwargs)

    # Note: id is used to differentiate objects on the same line.
    self.insert_trace = create_table(self.basename + "_trace", "filename", "lineno", "id", *self.traced_information)

  def start_tree(self, *args, **kwargs):
    super(BaseStaticTraceFix, self).start_tree(*args, **kwargs)

    self.ids = defaultdict(int)


class BaseDynamicTraceFix(BaseStaticTraceFix):
  def __init__(self, *args, **kwargs):
    super(BaseDynamicTraceFix, self).__init__(*args, **kwargs)

    def f(filename, lineno, id_, *args, **kwargs):
      result, values = self._dynamic_trace(*args, **kwargs)
      self.insert_trace(filename, lineno, id_, *values)
      return result
    f.__name__ = self.basename + "_traced"
    tracing_functions.append(f)

    self.insert_modified = create_table(self.basename + "_modified", "filename", "lineno", "id")

  def traced_call(self, node, children):
    # Important: every node inside "children" should be cloned before this
    # function is called.

    first_child = children[0]
    if node.prefix == first_child.prefix:
      first_child.prefix = ''

    filename = self.filename
    lineno = node.get_lineno()
    id_ = self.ids[lineno]

    new_node = lib2to3.fixer_util.Call(
      lib2to3.fixer_util.Name(self.basename + "_traced"),
      args=[
        lib2to3.fixer_util.String('"%s"' % filename),
        lib2to3.fixer_util.Comma(),
        lib2to3.fixer_util.Number(lineno),
        lib2to3.fixer_util.Comma(),
        lib2to3.fixer_util.Number(id_),
        lib2to3.fixer_util.Comma(),
      ] + children,
      prefix=node.prefix)

    self.ids[lineno] += 1

    self.insert_modified(filename, lineno, id_)

    return new_node


class BaseSupportFix(BaseFix):
  def __init__(self, *args, **kwargs):
    super(BaseSupportFix, self).__init__(*args, **kwargs)

    # Note: id is used to differentiate objects on the same line.
    self.insert_support = create_table(self.basename + "_support", "filename", "lineno", "id", "status")

  def analyze_data(self, filename, lineno, id_):
    class_name = "Fix%sTrace" % self.basename.capitalize()
    traced_information = getattr(
      __import__("my2to3.fixes.fix_%s_trace" % self.basename, fromlist=[class_name]),
      class_name
    ).traced_information

    data = get_data(
      self.basename + "_trace",
      traced_information,
      dict(filename=filename, lineno=lineno, id=id_)
    )

    if data:
      try:
        results, status = self._analyze_data()
      except Exception:
        results, status = False, "unknown"
    else:
      results, status = False, "no data"

    self.insert_support(filename[1:-1], lineno, id_, status)
    return results
